% FLENG compiler - IR -> assembly stage

init_asm(Infile, Out) :-
    fleng_version(V),
    format_to_atom(C, 'GENERATED BY fl2x64 [~a] FROM ~a\n', [V, Infile]),
    d_setup(Out),
    d_comment(C, Out).

asm([], _).
asm([Instr|Asm], Out) :-
    asm(Instr, Out),
    asm(Asm, Out).

%asm(X, _) :- dbg(X), fail.
asm(above, Out) :- xcall(fl_above, Out).
asm(addref, Out) :- ra_inc(a, Out).
asm(addrefx, Out) :- xcall(fl_addref, Out).
asm(align_stack, Out) :- sp_align(Out).
asm(asm(L), Out) :- emit('~a\n', [L], Out).
asm(assign, Out) :- xcall_d(assign0, Out).
asm(atom(I, _), Out) :- a_load(atoms, I, a, Out).
asm(atomic, Out) :- xcall(fl_atomic, Out).
asm(atoms(N), Out) :- 
    d_section('.bss', Out),
    d_label(atoms, Out),
    d_wreserve(N, Out),
    d_equ('S', N, 0, Out),
    d_section('.data', Out).
asm(below, Out) :- xcall(fl_below, Out).
asm(bind(I), Out) :- 
    r_move(a, e(I), Out),
    ra_inc(a, Out).
asm(cdr, Out) :- r_load(a, 2, a, Out).
asm(check_atom, Out) :- xcall(fl_check_atom, Out).
asm(check_cell, Out) :- xcall(fl_check_cell, Out).
asm(check_integer, Out) :- xcall(fl_check_integer, Out).
asm(check_list, Out) :- xcall(fl_check_list, Out).
asm(check_tuple, Out) :- xcall(fl_check_tuple, Out).
asm(check_number, Out) :- xcall(fl_check_number, Out).
asm(check_module, Out) :- xcall(fl_check_module, Out).
asm(check_var, Out) :- xcall(fl_check_var, Out).
asm(close, _).
asm(code_label(N/A, I), Out) :-
    format_to_atom(C, 'label: ~a/~d', [N, A]),
    d_comment(C, Out),
    la_loadm('P', I, a, Out).
asm(compute(+), Out) :- xcall(fl_add, Out).
asm(compute(-), Out) :- xcall(fl_sub, Out).
asm(compute(*), Out) :- xcall(fl_mul, Out).
asm(compute(/), Out) :- xcall(fl_div, Out).
asm(compute(mod), Out) :- xcall(fl_mod, Out).
asm(compute(and), Out) :- r_and(p(0), p(1), a, Out).
asm(compute(or), Out) :- r_or(p(0), p(1), a, Out).
asm(compute(xor), Out) :- r_xor(p(0), p(1), a, Out).
asm(compute(shl), Out) :- xcall(fl_shl, Out).
asm(compute(shr), Out) :- xcall(fl_shr, Out).
asm(compute(=), Out) :- xcall(fl_equal, Out).
asm(compute(<), Out) :- xcall(fl_less, Out).
asm(compute(>), Out) :- xcall(fl_greater, Out).
asm(compute(sametype), Out) :- xcall(fl_sametype, Out).
asm(compute(**), Out) :- xcall(fl_pow, Out).
asm(compute(sqrt), Out) :- xcall(fl_sqrt, Out).
asm(compute(sin), Out) :- xcall(fl_sin, Out).
asm(compute(cos), Out) :- xcall(fl_cos, Out).
asm(compute(tan), Out) :- xcall(fl_tan, Out).
asm(compute(atan), Out) :- xcall(fl_atan, Out).
asm(compute(log), Out) :- xcall(fl_loge, Out).
asm(compute(exp), Out) :- xcall(fl_exp, Out).
asm(compute(integer), Out) :- xcall(fl_int, Out).
asm(compute(real), Out) :- xcall(fl_float, Out).
asm(compute(truncate), Out) :- xcall(fl_truncate, Out).
asm(compute(round), Out) :- xcall(fl_round, Out).
asm(compute(floor), Out) :- xcall(fl_floor, Out).
asm(compute(ceiling), Out) :- xcall(fl_ceiling, Out).
asm(compute(sign), Out) :- xcall(fl_sign, Out).
asm(compute(rnd), Out) :- xcall(fl_rnd, Out).
asm(compute(abs), Out) :- xcall(fl_abs, Out).
asm(compute(real_integer_part), Out) :- xcall(fl_fip, Out).
asm(compute(real_fractional_part), Out) :- xcall(fl_ffp, Out).
asm(compute(max), Out) :- xcall(fl_max, Out).
asm(compute(min), Out) :- xcall(fl_min, Out).
asm(data(L), Out) :- 
    d_label('L', L, Out),
    d_wdata([0], Out).
asm(decons, Out) :-
    r_load(a, 2, x(0), Out),
    r_push(x(0), Out),
    r_load(a, 1, a, Out).
asm(defatom(I, A), Out) :-
    string_length(A, Len),
    d_align(halfword, Out),
    d_bdata([0, 0], Out),
    d_label('S', I, Out),
    atom_codes(A, C),
    d_bdata([Len|C], Out),
    d_bdata([0], Out),
    NA is I + 1,
    d_align(word, Out),
    format_to_atom(LB, 'S~d', [NA]),
    d_wdata([LB], Out).
asm(define(S, A, I, _, N), Out) :-
    % assumes: G, T, FP, SP are set
    assertz(pdef_name(I, S)),
    format_to_atom(C, '~a/~d', [N, A]),
    d_comment(C, Out),
    d_align(code, Out),
    d_label('P', I, Out),
    emit_entry_point(N/A, Out),
    la_load('D', I, x(0), Out),
    r_store(x(0), g, 5, Out),   % goal->info
    emit_entry(A, Out).
asm(deref, Out) :- xcall(fl_deref, Out).
asm(drop(I), Out) :- 
    r_move(e(I), x(0), Out),
    xcall(fl_unref, Out).
asm(droparg(I), Out) :- 
    I2 is I + 1,
    r_load(g, I2, x(0), Out),
    xcall(fl_unref, Out).
asm(dropgoal, Out) :- xcall(fl_dropgoal, Out).
asm(end, Out) :- 
    d_finalize(Out),
    module_info(N, SMod, _),
    d_section('.data', Out),
    d_label(initd, Out),
    d_wdata([0], Out),
    d_label(module, Out),
    d_wdata([0, N, SMod, atoms, 'S0', ratoms, 'PI0'], Out),
    emit_pdefs(Out),
    emit_info_strings(Out),
    d_equ('PI', 0, 0, Out),
    d_comment('END OF FILE', Out).
asm(end_atoms, Out) :- d_wdata([0], Out).
asm(end_def, Out) :- d_endfunction(Out).
asm(error, Out) :- 
    r_move(t, c, Out),
    xcall(fl_abort, Out).
asm(fail(A), Out) :-
    i_load(A, p(0), Out),
    r_move(t, c, Out),
    xcall(fl_fail, Out).
asm(force(I), Out) :- 
    xcall(fl_force, Out),
    j_if_carry('L', I, Out).
asm(foreign_call(N, A), Out) :- 
    format_to_atom(Op, '~a_~d', [N, A]),
    emit_external(Op, Out),
    c_prepare_foreign_call(A, Out),
    r_move(t, c, Out),
    xcall(Op, Out),
    c_finalize_foreign_call(A, Out).
asm(getarg(I), Out) :- 
    I2 is I + 1,
    r_load(g, I2, a, Out).
asm(getenv(I), Out) :- r_move(e(I), a, Out).
asm(getregister(R), Out) :- r_move_r(R, a, Out).
asm(guard(I), Out) :- 
    j_r_compare_if_not_equal(p(0), p(1), 'L', I, Out).
asm(idle(I), Out) :- 
    la_load('L', I, p(0), Out),
    r_move(t, c, Out),
    xcall(fl_idle, Out).
asm(getmodule, Out) :- 
    la_load(module, a, Out),
    i_or(1, a, Out).
asm(init, Out) :- emit_module_init(none, Out).
asm(init(I), Out) :- emit_module_init(I, Out).  % init with startup goal
asm(invoke_goal, Out) :- 
    r_move(g, n, Out),
    i_mask(a, '~1', Out),
    r_store(a, n, 0, Out).
asm(invoke_jump, Out) :- xcall(fl_invoke_jump, Out).
asm(if_false(I), Out) :- 
    j_a_compare_if_not_equal(a, 'fl_false', 'L', I, Out).
asm(if_true(I), Out) :- 
    j_a_compare_if_not_equal(a, 'fl_true', 'L', I, Out).
asm(isatom(A, L, LF, _), Out) :- 
    a_load(atoms, A, p(1), Out),
    la_load('L', L, p(0), Out),
    xcall(fl_isval, Out),
    j_if_carry('L', LF, Out).
asm(isint(N, L, LF), Out) :- 
    X is (N << 1) \/ 1,
    i_load(X, p(1), Out),
    la_load('L', L, p(0), Out),
    xcall(fl_isval, Out),
    j_if_carry('L', LF, Out).
asm(isfloat(N, L, LF), Out) :- 
    f_load(N, Out),
    la_load('L', L, p(0), Out),
    xcall(fl_isfloat, Out),
    j_if_carry('L', LF, Out).
asm(islist(L, LF), Out) :- 
    la_load('L', L, p(0), Out),
    xcall(fl_islist, Out),
    j_if_carry('L', LF, Out).
asm(islist_n(L), Out) :- 
    xcall(fl_islist_t, Out),
    j_if_carry('L', L, Out),
    r_load(a, 2, x(0), Out),
    r_push(x(0), Out),
    r_load(a, 1, a, Out).
asm(islist_t(L), Out) :- 
    xcall(fl_islist_t, Out),
    j_if_carry('L', L, Out).
asm(istuple(N, L, LF), Out) :-
    i_load(N, p(1), Out),
    la_load('L', L, p(0), Out),
    xcall(fl_istuple, Out),
    j_if_carry('L', LF, Out).
asm(jump(L), Out) :- j_always('L', L, Out).
asm(label(L), Out) :- d_label('L', L, Out).
asm(literal(L), Out) :- la_load('C', L, a, Out).
asm(match, Out) :- xcall(fl_match, Out).
asm(mkint(I), Out) :- 
    X is (I << 1) \/ 1,
    i_load(X, a, Out).
asm(mkfloat(N), Out) :- 
    f_load(N, Out),
    xcall(fl_mkfloat, Out).
asm(mklist, Out) :- xcall(fl_mklist, Out).
asm(mktuple(I, Arity, _), Out) :- 
    i_load(Arity, x(0), Out),
    a_load(atoms, I, x(1), Out),
    xcall(fl_mktuple, Out).
asm(mktuple(Arity), Out) :- 
    i_load(Arity, x(0), Out),
    xcall(fl_mktuple0, Out).
asm(mkvar, Out) :- xcall(fl_mkvar, Out).
asm(module(N, SMod, Name), Out) :-
    assertz(module_info(N, SMod, Name)),
    d_section('.text', Out).
asm(not_suspending(L), Out) :- 
    r_load(t, 18, x(0), Out),   % tcb->sstack
    j_compare_if_equal(x(0), t, 1, 'L', L, Out). % tcb->sstackp
asm(open_goal(I, N, A), Out) :-
    integer(I),
    !,
    xcall(fl_opengoal, Out),
    format_to_atom(C, 'open: ~a/~d', [N, A]),
    d_comment(C, Out),
    la_load('P', I, a, Out),
    r_store(a, n, 0, Out).
asm(open_goal(I, N, A), Out) :-
    xcall(fl_opengoal, Out),
    emit_external(I, Out),
    format_to_atom(C, 'open: ~a/~d', [N, A]),
    d_comment(C, Out),
    lag_load(I, a, Out),
    r_store(a, n, 0, Out).
asm(pop, Out) :- r_pop(a, Out).
asm(push, Out) :- r_push(a, Out).
asm(putarg(I), Out) :- 
    I2 is I + 1,
    r_store(a, n, I2, Out).
asm(putenv(I), Out) :- 
    r_move(a, e(I), Out),
    xcall(fl_addref, Out).
asm(putenvu(I), Out) :- r_move(a, e(I), Out).
asm(putasmparam(I), Out) :- r_move(a, p(I), Out).
asm(putparam(I), Out) :- r_move(a, p(I), Out).
asm(puttailarg, Out) :- xcall(fl_puttailarg, Out).
asm(reify, Out) :- xcall(fl_reify, Out).
asm(reset, Out) :-
    r_load(t, 18, x(0), Out),   % tcb->sstack
    r_store(x(0), t, 1, Out).   % tcb->sstackp
asm(resolve(L, M, N, A), Out) :- 
    la_load('L', L, p(3), Out),
    i_load(A, p(0), Out),
    a_load(atoms, M, p(1), Out),
    a_load(atoms, N, p(2), Out),
    xcall(fl_resolve, Out).
asm(restore(R), Out) :- r_pop(p(R), Out).
asm(save(R), Out) :- r_push(p(R), Out).
asm(setarg(I), Out) :- 
    I2 is I + 1,
    ra_add(g, I2, p(1), Out),
    xcall(fl_replace, Out).
asm(suspend, Out) :- 
    r_move(t, c, Out),
    xcall_d(suspend, Out).
asm(suspending(L), Out) :- 
    r_load(t, 18, p(0), Out),   % tcb->sstack
    j_compare_if_not_equal(p(0), t, 1, 'L', L, Out). % tcb->sstackp
asm(switch_atomic(L), Out) :- 
    j_if_test(a, 3, 'L', L, Out).
asm(switch_list(L), Out) :- 
    xcall(fl_islist_t, Out),
    j_if_carry('L', L, Out).
asm(switch_tuple(L), Out) :- 
    xcall(fl_istuple_t, Out),
    j_if_carry('L', L, Out).
asm(tailgoal(I, N, A), Out) :- 
    integer(I),
    !,
    format_to_atom(C, 'tail: ~a/~d', [N, A]),
    d_comment(C, Out),
    la_load('P', I, p(1), Out),
    xcall(fl_tailgoal, Out).
asm(tailgoal(I, N, A), Out) :- 
    emit_external(I, Out),
    format_to_atom(C, 'tail: ~a/~d', [N, A]),
    d_comment(C, Out),
    lag_load(I, p(1), Out),
    xcall(fl_tailgoal, Out).
asm(tailjump, Out) :- xcall(fl_tail_jump, Out).
asm(test_float, Out) :-
    i_load(0x02000000, x(0), Out),    % FL_FLOAT_TAG
    xcall(fl_test_tag, Out).
asm(test_integer, Out) :- xcall(fl_test_integer, Out).
asm(test_number, Out) :- xcall(fl_test_number, Out).
asm(test_atom, Out) :- xcall(fl_test_atom, Out).
asm(test_list, Out) :- 
    i_load(0x01000000, x(0), Out),    % FL_LIST_TAG
    xcall(fl_test_tag, Out).
asm(test_module, Out) :- 
    i_load(0x0a000000, x(0), Out),    % FL_MODULE_TAG
    xcall(fl_test_tag, Out).
asm(test_port, Out) :- 
    i_load(0x05000000, x(0), Out),    % FL_PORT_TAG
    xcall(fl_test_tag, Out).
asm(test_tuple, Out) :- xcall(fl_test_tuple, Out).
asm(test_var, Out) :- xcall(fl_test_var, Out).
asm(test_remote, Out) :- xcall(fl_test_remote, Out).
asm(try_else(L), Out) :- 
    d_endfunction(Out),
    d_align(halfword, Out),
    d_label('L', L, Out),
    sp_reset(Out),
    r_load(t, 5, sp, Out).  % tcb->stackbase
asm(unify, Out) :- xcall_d(unify, Out).
asm(unify_safe, Out) :- xcall(fl_unify_safe, Out).
asm(unify_checked, Out) :- xcall_d(unify_checked, Out).
asm(unref(I), Out) :- 
    %XXX could dec refcount and test inline, as value is always a cell
    r_move(p(I), x(0), Out),
    xcall(fl_unref, Out).
asm(yield, Out) :- 
    r_move(t, c, Out),
    xcall(fl_yield, Out).

xcall_d(Call, Out) :-
    debug_info,
    !,
    atom_concat('fl_d_', Call, Call2),
    xcall(Call2, Out).
xcall_d(Call, Out) :- 
    atom_concat('fl_', Call, Call2),
    xcall(Call2, Out).

emit_module_init(Init, Out) :-
    module_info(_, _, Name),
    format_to_atom(IP, 'fl_~a_init', [Name]),
    d_global(IP, Out),
    d_align(code, Out),
    dg_label(IP, Out),
    d_function_entry(Out),
    a_load(initd, 0, a, Out),
    j_if_not_test(a, 1, 'Li', Out),
    la_inc(initd, Out),
    la_load(module, c, Out),
    xcall(fl_init_module, Out),
    emit_uses(Out),
    emit_init_goal(Init, Out),
    d_label('Li', Out),
    d_function_exit(Out),
    g_return(Out),
    d_endfunction(Out),
    emit_cmdline_init(Out).

emit_cmdline_init(Out) :-
    module_info(_, _, ''),
    (cmdline_init(Cmdline); Cmdline = [0]),
    d_section('.data', Out),
    d_global(fl_cmdline, Out),
    dg_label(fl_cmdline, Out),
    d_bdata(Cmdline, Out),
    d_bdata([0], Out).
emit_cmdline_init(_).

emit_init_goal(none, _) :- !.
emit_init_goal(Init, Out) :-
    la_load('PD', Init, c, Out),
    xcall(fl_init_goal, Out).

emit_pdefs(Out) :- emit_pdefs(0, Out).

emit_pdefs(N, Out) :- 
    pdef(Name/Arity, N),
    !,
    pdef_name(N, S),
    d_label('PD', N, Out),
    format_to_atom(PL, 'P~d', [N]),
    export_flag(Name/Arity, XF),
    d_wdata([S, Arity, XF, PL], Out),
    N2 is N + 1,
    emit_pdefs(N2, Out).
emit_pdefs(_, Out) :- 
    d_wdata([0], Out).

export_flag(NA, 1) :- exported(NA), !.
export_flag(_, 1) :- \+exported(_), !.
export_flag(_, 0).

emit_entry_point(N/A, Out) :-
    get_module_name(M),
    entry_point(M:N/A, Label),
    !,
    d_global(Label, Out),
    dg_label(Label, Out).
emit_entry_point(_, _).

emit_entry(A, Out) :-
    debug_info,
    !,
    i_load(A, p(0), Out),
    xcall(fl_entry, Out).
emit_entry(_, _).

emit_info_strings(Out) :-
    pdef(N/A, I),
    d_label('D', I, Out),
    format_to_atom(NA, '~q/~d', [N, A]),
    atom_codes(NA, NC),
    d_bdata(NC, Out),
    d_bdata([0], Out),
    fail.
emit_info_strings(_).

emit_uses(Out) :-
    uses(Mod),
    format_to_atom(IName, 'fl_~a_init', [Mod]),
    emit_external(IName, Out),
    a_call(IName, Out),
    fail.
emit_uses(_).

xcall(Name, Out) :-
    emit_external(Name, Out),
    a_call(Name, Out).

dump_resolved_atoms(Out) :-
    d_label(ratoms, Out),
    fail.
dump_resolved_atoms(Out) :-
    symbol_loc(SL, _),
    format_to_atom(L, 'SL~d', [SL]),
    d_wdata([L], Out),
    fail.
dump_resolved_atoms(Out) :- d_wdata([0], Out).

% preallocated literal dumping
dump_literal(N, L, Out) :-
    float(N), 
    d_align(float, Out),
    d_label('C', L, Out),
    d_wdata([0x02ffffff], Out),
    d_fdata([N], Out),
    d_wdata([1], Out).
dump_literal([X|Y], L, Out) :-
    !,
    d_align(word, Out),
    d_label('C', L, Out),
    d_wdata([0x01ffffff], Out),
    dump_sub_literal(X, Out),
    dump_sub_literal(Y, Out).
dump_literal(functor(N, A, AL), L, Out) :-
    Tag is 0x80ffffff \/ (A << 24),
    d_align(word, Out),
    d_label('C', L, Out),
    format_to_atom(CL, 'C~d', [AL]),
    d_wdata([Tag], Out),
    register_symbol_loc(N, SL),
    d_label('SL', SL, Out),
    d_wdata([N, CL], Out).
dump_literal(tuple(A, HL, AL), L, Out) :-
    Tag is 0x80ffffff \/ (A << 24),
    d_align(word, Out),
    d_label('C', L, Out),
    format_to_atom(CL, 'C~d', [AL]),
    d_wdata([Tag], Out),
    dump_sub_literal(HL, Out),
    d_wdata([CL], Out).
dump_literal(tuple0(HL), L, Out) :-
    Tag is 0x80ffffff,
    d_align(word, Out),
    d_label('C', L, Out),
    d_wdata([Tag], Out),
    dump_sub_literal(HL, Out),
    dump_sub_literal([], Out).

dump_sub_literal(X, Out) :-
    integer(X), 
    !,
    N is (X << 1) \/ 1,
    d_wdata([N], Out).
dump_sub_literal([], Out) :-
    % for swipl
    !,
    r_atom('[]', A),
    register_symbol_loc(A, SL),
    d_label('SL', SL, Out),
    d_wdata([A], Out).
dump_sub_literal(X, Out) :-
    atom(X), 
    !,
    r_atom(X, A),
    register_symbol_loc(A, SL),
    d_label('SL', SL, Out),
    d_wdata([A], Out).
dump_sub_literal(label(L), Out) :-
    format_to_atom(CL, 'C~d', [L]),
    d_wdata([CL], Out).
