% FLENG compiler - clause indexing

compile_group(N/A, I, EL, G, Lc1, Lc, Out) :-
    print_progress(N, A),
    register_atom(N, S),
    asm([define(S, A, I, EL, N), reset], Out),
    partition_on_first_arg(G, ACs, LCs, TCs, OCs),
    select_indexing_method(ACs, LCs, TCs, OCs, G, Lc1, Lc2, Out),
    label(Lc2, Lc, Lf),
    asm([not_suspending(Lf), suspend, yield, label(Lf)], Out),
    asm([fail(A), end_def], Out).

print_progress(N, A) :-
    print_progress,
    format('~a/~d\n', [N, A]).
print_progress(_, _).

select_indexing_method(ACs, LCs, TCs, OCs, _, Lc1, Lc, Out) :-
    length(ACs, Len1),
    length(LCs, Len2),
    length(TCs, Len3),
    Len1 + Len2 + Len3 > 2,
    !,
    compile_indexed(ACs, LCs, TCs, OCs, Lc1, Lc, Out).
select_indexing_method(_, _, _, _, All, Lc1, Lc, Out) :-
    compile_unindexed(All, none, yes, Lc1, Lc, Out).

partition_on_first_arg([], [], [], [], []).
partition_on_first_arg([t((H :- T), Vs, Ss)|G], ACs, LCs, TCs, OCs) :-
    H =.. [_, Arg|_],
    partition_on_first_arg(Arg, t((H :-T), Vs, Ss), G, ACs, LCs, TCs, OCs).
partition_on_first_arg(G, [], [], [], G).

partition_on_first_arg('$VAR'(_), C, G, ACs, LCs, TCs, [C|OCs]) :-
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg('!'(Arg), C, G, ACs, LCs, TCs, OCs) :-
    partition_on_first_arg(Arg, C, G, ACs, LCs, TCs, OCs).
partition_on_first_arg(V, C, G, [C|ACs], LCs, TCs, OCs) :-
    atomic(V), \+float(V),
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg([_|_], C, G, ACs, [C|LCs], TCs, OCs) :-
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg(F, C, G, ACs, LCs, [C|TCs], OCs) :-
    compound(F),
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).
partition_on_first_arg(_, C, G, ACs, LCs, TCs, [C|OCs]) :-
    partition_on_first_arg(G, ACs, LCs, TCs, OCs).    

compile_unindexed([], Fail, _, Lc, Lc, Out) :-
    emit_fail_label(Fail, Out).
compile_unindexed([t(Term, _, Ss)|G], Fail, ForceArg1, Lc1, Lc, Out) :-
    emit_fail_label(Fail, Out),
    label(Lc1, Lc2, Fail2),
    compile_clause(Term, Ss, ForceArg1, Fail2, Lc2, Lc3, Out),
    compile_unindexed(G, Fail2, yes, Lc3, Lc, Out).

emit_fail_label(none, _).
emit_fail_label(Fail, Out) :-
    number(Fail), 
    asm(try_else(Fail), Out).

% force first arg and dispatch according to type

compile_indexed(ACs, LCs, TCs, OCs, Lc1, Lc, Out) :-
    label(Lc1, Lc2, Lo),
    asm([getarg(0), force(Lo), setarg(0)], Out),
    compile_indexed2(ACs, LCs, TCs, OCs, none, Lo, Lc2, Lc, Out).

compile_indexed2([C|ACs], LCs, TCs, OCs, Fail, Lo, Lc1, Lc, Out) :-
    emit_fail_label(Fail, Out),
    label(Lc1, Lc2, Fail2),
    asm([getarg(0), switch_atomic(Fail2)], Out),
    compile_unindexed([C|ACs], none, no, Lc2, Lc3, Out),
    compile_indexed2([], LCs, TCs, OCs, Fail2, Lo, Lc3, Lc, Out).
compile_indexed2([], [C|LCs], TCs, OCs, Fail, Lo, Lc1, Lc, Out) :-
    emit_fail_label(Fail, Out),
    label(Lc1, Lc2, Fail2),
    asm([getarg(0), switch_list(Fail2)], Out),
    compile_unindexed([C|LCs], none, no, Lc2, Lc3, Out),
    compile_indexed2([], [], TCs, OCs, Fail2, Lo, Lc3, Lc, Out).
compile_indexed2([], [], [C|TCs], OCs, Fail, Lo, Lc1, Lc, Out) :-
    emit_fail_label(Fail, Out),
    asm([getarg(0), switch_tuple(Lo)], Out),
    compile_unindexed([C|TCs], none, no, Lc1, Lc2, Out),
    compile_indexed2([], [], [], OCs, none, Lo, Lc2, Lc, Out).
compile_indexed2([], [], [], OCs, Fail, Lo, Lc1, Lc, Out) :-
    emit_fail_label(Fail, Out),
    asm(label(Lo), Out),
    compile_unindexed(OCs, none, no, Lc1, Lc, Out).
